// Spartan Schema
// Copyright © 2021-2023 Adam Nelson <adam@nels.onl>
// Distributed under the Blue Oak Model License
import isPlainObject from './isPlainObject.js';
import { pathToString } from './path.js';
function isSchemaType(schema, refs, location, errors) {
    if (schema === null) {
        return true;
    }
    else if (Array.isArray(schema)) {
        if (schema.length < 2) {
            errors &&
                errors.push({
                    message: 'Array types must have at least 2 elements',
                    location,
                });
            return false;
        }
        if (typeof schema[0] !== 'string') {
            errors &&
                errors.push({
                    message: 'Array types must start with a string',
                    location: [...location, 0],
                });
            return false;
        }
        switch (schema[0]) {
            case 'enum': {
                let result = true;
                for (let i = 1; i < schema.length; i++) {
                    switch (typeof schema[i]) {
                        case 'boolean':
                        case 'number':
                        case 'string':
                            continue;
                        default:
                            if (schema[i] === null) {
                                continue;
                            }
                            result = false;
                            errors &&
                                errors.push({
                                    message: '"enum" type elements must be boolean, number, string, or null',
                                    location: [...location, i],
                                });
                    }
                }
                return result;
            }
            case 'tuple':
            case 'array':
            case 'oneof': {
                let result = true;
                for (let i = 1; i < schema.length; i++) {
                    if (!isSchemaType(schema[i], refs, [...location, i], errors)) {
                        result = false;
                        if (!errors) {
                            break;
                        }
                    }
                }
                return result;
            }
            case 'dictionary':
                if (schema.length !== 2) {
                    errors &&
                        errors.push({
                            message: '"dictionary" type must have exactly 2 elements',
                            location,
                        });
                    return false;
                }
                return isSchemaType(schema[1], refs, [...location, 1], errors);
            case 'ref':
                if (schema.length !== 2) {
                    errors &&
                        errors.push({
                            message: '"ref" type must have exactly 2 elements',
                            location,
                        });
                    return false;
                }
                if (typeof schema[1] !== 'string') {
                    errors &&
                        errors.push({
                            message: '"ref" type label must be a string',
                            location: [...location, 1],
                        });
                    return false;
                }
                if (!(schema[1] in refs)) {
                    errors &&
                        errors.push({
                            message: `"ref" type refers to nonexistent variable ${JSON.stringify(schema[1])}`,
                            location: [...location, 1],
                        });
                    return false;
                }
                return true;
            case 'optional':
                errors &&
                    errors.push({
                        message: '"optional" type is only allowed as an object key',
                        location,
                    });
                return false;
            default:
                errors &&
                    errors.push({
                        message: `Unknown array type: ${JSON.stringify(schema[0])}`,
                        location: [...location, 0],
                    });
                return false;
        }
    }
    else if (isPlainObject(schema)) {
        let result = true;
        for (const [k, v] of Object.entries(schema)) {
            let s;
            if (Array.isArray(v) && v.length > 0 && v[0] === 'optional') {
                if (v.length !== 2) {
                    result = false;
                    if (errors) {
                        errors &&
                            errors.push({
                                message: '"optional" type must have exactly 2 elements',
                                location: [...location, k],
                            });
                        continue;
                    }
                    else {
                        break;
                    }
                }
                s = v[1];
            }
            else {
                s = v;
            }
            if (!isSchemaType(s, refs, [...location, k], errors)) {
                result = false;
                if (!errors) {
                    break;
                }
            }
        }
        return result;
    }
    else if (typeof schema === 'string') {
        switch (schema) {
            case 'null':
            case 'boolean':
            case 'integer':
            case 'float':
            case 'number':
            case 'string':
            case 'date':
            case 'binary':
                return true;
            default:
                errors &&
                    errors.push({
                        message: `Unknown scalar type: ${JSON.stringify(schema)}`,
                        location,
                    });
                return false;
        }
    }
    errors &&
        errors.push({
            message: `Literal value outside "enum": ${JSON.stringify(schema)}`,
            location,
        });
    return false;
}
/**
 * A type predicate that checks whether `schema` is a valid Spartan Schema.
 *
 * `errors` is a mutable array of `{ message, location }` pairs; if it is
 * present and `isSchema` returns false, it will be populated with a list of
 * parsing errors.
 */
export function isSchema(schema, errors) {
    if (!isPlainObject(schema)) {
        errors &&
            errors.push({ message: 'Schema must be an object', location: [] });
        return false;
    }
    let result = true;
    if ('spartan' in schema) {
        if (schema['spartan'] !== 1) {
            if (errors) {
                errors &&
                    errors.push({
                        message: 'Schema property "spartan" must be 1, if present',
                        location: ['spartan'],
                    });
                result = false;
            }
            else {
                return false;
            }
        }
    }
    let refs = {};
    if ('let' in schema) {
        if (isPlainObject(schema['let'])) {
            refs = schema['let'];
            for (const [k, v] of Object.entries(refs)) {
                if (!isSchemaType(v, refs, ['let', k], errors)) {
                    result = false;
                    if (!errors) {
                        return false;
                    }
                }
            }
        }
        else if (errors) {
            errors &&
                errors.push({
                    message: 'Schema property "let" must be an object, if present',
                    location: ['let'],
                });
            result = false;
        }
        else {
            return false;
        }
    }
    if ('schema' in schema) {
        return isSchemaType(schema['schema'], refs, ['schema'], errors) && result;
    }
    else {
        errors &&
            errors.push({
                message: 'Schema must have a "schema" property',
                location: [],
            });
        return false;
    }
}
export function validationErrorToString({ dataPath, schemaPath, message, children }, maxLength = 8, maxDepth = 4) {
    return `${pathToString(dataPath)}: ${message} (rule ${pathToString(schemaPath)})${maxDepth && children?.length
        ? children.slice(0, maxLength).map((cs) => '\n- ' + cs.slice(0, maxLength).map((c) => validationErrorToString(c, maxLength, maxDepth - 1).replaceAll('\n', '\n  ')).join('\n  ') + (cs.length > maxLength ? `\n  (...and ${cs.length - maxLength} more)` : '')).join('') +
            (children.length > maxLength ? `\n- (...and ${children.length - maxLength} more)` : '')
        : ''}`;
}
/**
 * A detailed error thrown by `assertMatchesSchema` when schema validation
 * fails. Includes the schema, the JSON that didn't match, and the list of
 * validation errors, complete with JSONPath locations in both the schema and
 * the JSON value.
 */
export class SchemaAssertionError extends Error {
    constructor(json, schema, validationErrors, initialMessage = 'Schema validation failed') {
        super();
        Object.defineProperty(this, "json", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: json
        });
        Object.defineProperty(this, "schema", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: schema
        });
        Object.defineProperty(this, "validationErrors", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: validationErrors
        });
        Object.defineProperty(this, "initialMessage", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: initialMessage
        });
    }
    get message() {
        return `${this.initialMessage}\nJSON: ${JSON.stringify(this.json, null, 2)}\nValidation errors:\n${this.validationMessage}`;
    }
    get validationMessage() {
        const max = SchemaAssertionError.maxErrorsPerMessage;
        return this.validationErrors.slice(0, max).map((e) => validationErrorToString(e, max, 4)).join('\n') + (this.validationErrors.length > max
            ? `\n(...and ${this.validationErrors.length - max} more)`
            : '');
    }
}
Object.defineProperty(SchemaAssertionError, "maxErrorsPerMessage", {
    enumerable: true,
    configurable: true,
    writable: true,
    value: 8
});
function validateSchemaType(schema, refs, schemaPath = []) {
    if (schema === null) {
        return (value, dataPath) => value === null ? [] : [{ dataPath, schemaPath, message: 'expected null' }];
    }
    else if (Array.isArray(schema)) {
        switch (schema[0]) {
            case 'enum':
                return (value, dataPath) => {
                    for (let i = 1; i < schema.length; i++) {
                        if (value === schema[i]) {
                            return [];
                        }
                    }
                    return [{
                            dataPath,
                            schemaPath,
                            message: schema.length === 2
                                ? 'expected the exact value ' + JSON.stringify(schema[1])
                                : `expected one of ${schema.slice(1).map((x) => JSON.stringify(x)).join(', ')}`,
                        }];
                };
            case 'tuple': {
                const predicates = schema.slice(1).map((e, i) => validateSchemaType(e, refs, [...schemaPath, i + 1]));
                return (value, dataPath) => {
                    if (!Array.isArray(value)) {
                        return [{ dataPath, schemaPath, message: 'expected array' }];
                    }
                    else if (value.length !== predicates.length) {
                        return [{
                                dataPath,
                                schemaPath,
                                message: `expected array of exactly ${predicates.length} values, got ${value.length}`,
                            }];
                    }
                    return value.flatMap((v, i) => predicates[i](v, [...dataPath, i]));
                };
            }
            case 'array': {
                if (schema.length === 2) {
                    const e = validateSchemaType(schema[1], refs, [...schemaPath, 1]);
                    return (value, dataPath) => {
                        if (!Array.isArray(value)) {
                            return [{ dataPath, schemaPath, message: 'expected array' }];
                        }
                        return value.flatMap((v, i) => e(v, [...dataPath, i]));
                    };
                }
                else {
                    const predicates = schema.slice(1).map((e, i) => validateSchemaType(e, refs, [...schemaPath, i + 1]));
                    return (value, dataPath) => {
                        if (!Array.isArray(value)) {
                            return [{ dataPath, schemaPath, message: 'expected array' }];
                        }
                        else if (value.length < predicates.length - 1) {
                            return [{
                                    dataPath,
                                    schemaPath,
                                    message: `expected array of at least ${predicates.length - 1} values, got ${value.length}`,
                                }];
                        }
                        return value.flatMap((v, i) => predicates[i < predicates.length - 1 ? i : predicates.length - 1](v, [...dataPath, i]));
                    };
                }
            }
            case 'dictionary': {
                const e = validateSchemaType(schema[1], refs, [...schemaPath, 1]);
                return (value, dataPath) => {
                    if (!isPlainObject(value)) {
                        return [{ dataPath, schemaPath, message: 'expected object' }];
                    }
                    return Object.entries(value).flatMap(([k, v]) => e(v, [...dataPath, k]));
                };
            }
            case 'oneof': {
                const predicates = schema.slice(1).map((e, i) => validateSchemaType(e, refs, [...schemaPath, i + 1]));
                return (value, dataPath) => {
                    const children = [];
                    for (const p of predicates) {
                        const errors = p(value, dataPath);
                        if (!errors.length)
                            return [];
                        children.push(errors);
                    }
                    return [{
                            dataPath,
                            schemaPath,
                            message: `all ${predicates.length} options failed`,
                            children,
                        }];
                };
            }
            case 'ref':
                // Can't use refs[schema[1]] directly because of recursion
                return (value, dataPath) => refs[schema[1]]?.(value, dataPath) ?? [{
                        dataPath,
                        schemaPath,
                        message: `broken ref: schema has no let definition named ${JSON.stringify(schema[1])}`,
                    }];
        }
    }
    else if (isPlainObject(schema)) {
        const predicates = Object.entries(schema).map(([k, v]) => {
            const isOptional = Array.isArray(v) && v[0] === 'optional';
            const predicate = validateSchemaType(isOptional ? v[1] : v, refs, [
                ...schemaPath,
                k,
                ...isOptional ? [1] : [],
            ]);
            return (value, dataPath) => k in value
                ? predicate(value[k], [...dataPath, k])
                : (isOptional
                    ? []
                    : [{ dataPath, schemaPath, message: `missing required field ${JSON.stringify(k)}` }]);
        });
        return (value, dataPath) => {
            if (!isPlainObject(value)) {
                return [{ dataPath, schemaPath, message: 'expected object' }];
            }
            return predicates.flatMap((p) => p(value, dataPath));
        };
    }
    else {
        switch (schema) {
            case 'null':
                return (value, dataPath) => value === null ? [] : [{ dataPath, schemaPath, message: 'expected null' }];
            case 'boolean':
                return (value, dataPath) => typeof value === 'boolean' ? [] : [{ dataPath, schemaPath, message: 'expected boolean' }];
            case 'float':
            case 'number':
                return (value, dataPath) => typeof value === 'number' ? [] : [{ dataPath, schemaPath, message: 'expected number' }];
            case 'integer':
                return (value, dataPath) => typeof value === 'number' && Number.isInteger(value)
                    ? []
                    : [{ dataPath, schemaPath, message: 'expected integer' }];
            case 'string':
                return (value, dataPath) => typeof value === 'string' ? [] : [{ dataPath, schemaPath, message: 'expected string' }];
            case 'date':
                return (value, dataPath) => value instanceof Date ? [] : [{ dataPath, schemaPath, message: 'expected date' }];
            case 'binary':
                return (value, dataPath) => ArrayBuffer.isView(value) ? [] : [{ dataPath, schemaPath, message: 'expected binary' }];
        }
    }
    return (_, dataPath) => [{ dataPath, schemaPath, message: 'bad schema' }];
}
/**
 * A curried function that checks whether `value` matches `schema` and returns
 * a boolean.
 *
 * If `schema` is statically known at typechecking type (defined with
 * `as const`), then the function returned by `matchesSchema(schema)` will be a
 * type predicate.
 *
 * `errors` is a mutable array of `{ dataPath, schemaPath, message, children? }`
 * objects. If it is present and `matchesSchema` returns false, it will be
 * populated with a list of validation errors.
 */
export function matchesSchema({ schema, let: refs = {}, }) {
    const refValidators = {};
    for (const [k, v] of Object.entries(refs)) {
        refValidators[k] = validateSchemaType(v, refValidators, ['let', k]);
    }
    const validator = validateSchemaType(schema, refValidators, ['schema']);
    return (value, errors) => {
        const result = validator(value, []);
        if (result.length) {
            if (errors != null)
                errors.push(...result);
            return false;
        }
        return true;
    };
}
/**
 * A curried function that checks whether `value` matches `schema` and throws
 * a `SchemaAssertionError` if it doesn't.
 *
 * If `schema` is statically known at typechecking type (defined with
 * `as const`), then the function returned by `assertMatchesSchema(schema)` will
 * be a type assertion function.
 *
 * `message` is an optional message string to include in the thrown error.
 */
export function assertMatchesSchema(schema) {
    const refValidators = {};
    for (const [k, v] of Object.entries(schema.let ?? {})) {
        refValidators[k] = validateSchemaType(v, refValidators, ['let', k]);
    }
    const validator = validateSchemaType(schema.schema, refValidators, ['schema']);
    return (value, message) => {
        const result = validator(value, []);
        if (result.length) {
            throw new SchemaAssertionError(value, schema, result, message);
        }
    };
}
function typeZeroValue(schema, refs, history) {
    if (schema === null) {
        return null;
    }
    if (typeof schema === 'string') {
        switch (schema) {
            case 'null':
                return null;
            case 'boolean':
                return false;
            case 'float':
            case 'integer':
            case 'number':
                return 0;
            case 'string':
                return '';
            case 'date':
                return new Date(0);
            case 'binary':
                return new Uint8Array();
            default:
                return undefined;
        }
    }
    if (history.has(schema)) {
        throw new Error(`Cannot determine zero value of the infinite schema type ${JSON.stringify(schema)}`);
    }
    history.add(schema);
    if (Array.isArray(schema)) {
        switch (schema[0]) {
            case 'enum':
                return schema[1];
            case 'tuple':
                return schema.slice(1).map((n) => typeZeroValue(n, refs, history));
            case 'array':
                return schema.slice(1, schema.length - 1).map((n) => typeZeroValue(n, refs, history));
            case 'dictionary':
                return {};
            case 'oneof':
                return typeZeroValue(schema[1], refs, history);
            case 'ref':
                return typeZeroValue(refs[schema[1]], refs, history);
        }
    }
    else if (typeof schema === 'object') {
        return Object.entries(schema)
            .filter(([, v]) => !Array.isArray(v) || v[0] !== 'optional')
            .reduce((accum, [k, v]) => ({
            ...accum,
            [k]: typeZeroValue(v, refs, history),
        }), {});
    }
    return undefined;
}
/**
 * Returns the *zero value* of this schema's root type.
 *
 * | Type           | Zero value                                    |
 * | -------------- | --------------------------------------------- |
 * | `null`         | `null`                                        |
 * | boolean        | `false`                                       |
 * | integer, float | `0`                                           |
 * | string         | `""`                                          |
 * | binary         | `0`-length `Uint8Array`                       |
 * | date           | `new Date(0)` (Jan 1, 1970)                   |
 * | object         | object populated with properties' zero values |
 * | `oneof`        | zero value of first type                      |
 * | `enum`         | first enum value                              |
 * | `array`        | `[]`                                          |
 * | `tuple`        | array populated with elements' zero values    |
 * | `dictionary`   | `{}`                                          |
 *
 * This function typechecks the schema it receives. If it is passed a known
 * schema type `S` (defined `as const` in a Typescript file), then its return
 * type will be `MatchesSchema<S>`.
 *
 * May throw an exception if the schema type is infinitely recursive.
 */
export function zeroValue({ schema, let: refs = {}, }) {
    return typeZeroValue(schema, refs, new Set());
}
